/*
 * Copyright Fraunhofer ISE, energy & meteo Systems GmbH, and other contributors 2011
 *
 * This file is part of openIEC61850.
 * For more information visit http://www.openmuc.org 
 *
 * openIEC61850 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * openIEC61850 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with openIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openiec61850.scsm.mms.client;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import org.openiec61850.client.ClientACSI;
import org.openiec61850.common.ServiceError;
import org.openiec61850.common.model.DataSet;
import org.openiec61850.common.model.FunctionalConstraint;
import org.openiec61850.common.model.LogicalDevice;
import org.openiec61850.common.model.LogicalNode;
import org.openiec61850.common.model.ModelNode;
import org.openiec61850.common.model.ObjectReference;
import org.openiec61850.common.model.Server;
import org.openiec61850.common.model.datypes.basic.BOOLEAN;
import org.openiec61850.common.model.datypes.commonacsi.Timestamp;
import org.openiec61850.common.model.report.BufferedReportControlBlock;
import org.openiec61850.common.model.report.OptFields;
import org.openiec61850.common.model.report.Report;
import org.openiec61850.common.model.report.ReportControlBlock;
import org.openiec61850.common.model.report.ReportEntryData.ReasonCode;
import org.openiec61850.common.model.report.UnbufferedReportContrlBlock;
import org.openiec61850.jmms.iso.acse.AcseAssociation;
import org.openiec61850.jmms.iso.acse.ClientAcseSAP;
import org.openiec61850.jmms.mms.asn1.AccessResult;
import org.openiec61850.jmms.mms.asn1.AlternateAccess;
import org.openiec61850.jmms.mms.asn1.AlternateAccess.SubChoice.SubSeq_selectAlternateAccess;
import org.openiec61850.jmms.mms.asn1.AlternateAccess.SubChoice.SubSeq_selectAlternateAccess.SubChoice_accessSelection;
import org.openiec61850.jmms.mms.asn1.ConfirmedRequestPdu;
import org.openiec61850.jmms.mms.asn1.ConfirmedResponsePdu;
import org.openiec61850.jmms.mms.asn1.ConfirmedServiceRequest;
import org.openiec61850.jmms.mms.asn1.ConfirmedServiceResponse;
import org.openiec61850.jmms.mms.asn1.Data;
import org.openiec61850.jmms.mms.asn1.DefineNamedVariableListRequest;
import org.openiec61850.jmms.mms.asn1.DeleteNamedVariableListRequest;
import org.openiec61850.jmms.mms.asn1.DeleteNamedVariableListRequest.SubSeqOf_listOfVariableListName;
import org.openiec61850.jmms.mms.asn1.DeleteNamedVariableListResponse;
import org.openiec61850.jmms.mms.asn1.GetNameListRequest;
import org.openiec61850.jmms.mms.asn1.GetNameListRequest.SubChoice_objectScope;
import org.openiec61850.jmms.mms.asn1.GetNameListResponse;
import org.openiec61850.jmms.mms.asn1.GetNamedVariableListAttributesResponse;
import org.openiec61850.jmms.mms.asn1.GetVariableAccessAttributesRequest;
import org.openiec61850.jmms.mms.asn1.InformationReport;
import org.openiec61850.jmms.mms.asn1.InitRequestDetail;
import org.openiec61850.jmms.mms.asn1.InitiateRequestPdu;
import org.openiec61850.jmms.mms.asn1.InitiateResponsePdu;
import org.openiec61850.jmms.mms.asn1.ListOfVariableSeq;
import org.openiec61850.jmms.mms.asn1.MmsPdu;
import org.openiec61850.jmms.mms.asn1.ObjectClass;
import org.openiec61850.jmms.mms.asn1.ObjectName;
import org.openiec61850.jmms.mms.asn1.ObjectName.SubSeq_domain_specific;
import org.openiec61850.jmms.mms.asn1.ReadRequest;
import org.openiec61850.jmms.mms.asn1.ReadResponse;
import org.openiec61850.jmms.mms.asn1.ServiceError.SubChoice_errorClass;
import org.openiec61850.jmms.mms.asn1.UnconfirmedPDU;
import org.openiec61850.jmms.mms.asn1.UnconfirmedService;
import org.openiec61850.jmms.mms.asn1.VariableAccessSpecification;
import org.openiec61850.jmms.mms.asn1.VariableAccessSpecification.SubSeqOf_listOfVariable;
import org.openiec61850.jmms.mms.asn1.VariableSpecification;
import org.openiec61850.jmms.mms.asn1.WriteRequest;
import org.openiec61850.jmms.mms.asn1.WriteRequest.SubSeqOf_listOfData;
import org.openiec61850.jmms.mms.asn1.WriteResponse;
import org.openiec61850.scsm.mms.MmsAsn1ComplexTypeConverter;
import org.openiec61850.scsm.mms.MmsAsn1PrimitiveTypeConverter;
import org.openiec61850.scsm.mms.MmsConstants;
import org.openiec61850.scsm.mms.MmsReference;
import org.openmuc.jasn1.ber.BerByteArrayOutputStream;
import org.openmuc.jasn1.ber.types.BerBitString;
import org.openmuc.jasn1.ber.types.BerBoolean;
import org.openmuc.jasn1.ber.types.BerInteger;
import org.openmuc.jasn1.ber.types.BerNull;
import org.openmuc.jasn1.ber.types.string.BerVisibleString;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Stefan Feuerhahn
 * @author Tobias Weidelt
 */
public class MmsScsmClientAssociation implements ClientACSI {

	private static final BerInteger version = new BerInteger(new byte[] { (byte) 0x01, (byte) 0x01 });
	private static final BerBitString proposedParameterCbbBitString = new BerBitString(new byte[] { 0x03, 0x05,
			(byte) 0xf1, 0x00 });

	private static Logger logger = LoggerFactory.getLogger(MmsScsmClientAssociation.class);

	private AcseAssociation acseAssociation = null;
	private MmsScsmClientReceiver mmsScsmClientReceiver;

	private InetAddress serverAddress;
	private int serverPort;

	private BlockingQueue<MmsPdu> incomingResponses = new LinkedBlockingQueue<MmsPdu>();
	private BlockingQueue<MmsPdu> incomingReports = new LinkedBlockingQueue<MmsPdu>();

	Server model;

	private int invokeID = 0;

	// not being used:
	// private int negotiatedMaxServOutstandingCalling;
	// private int negotiatedMaxServOutstandingCalled;
	// private int negotiatedDataStructureNestingLevel;

	private int negotiatedMaxPduSize;

	private String[] mmsFCs = { "MX", "ST", "CO", "CF", "DC", "SP", "SG", "RP", "LG", "BR", "GO", "GS", "SV", "SE",
			"EX", "SR", "OR", "BL" };

	protected MmsScsmClientAssociation(String serverAccessPointReference, String authenticationParameter,
			ClientAcseSAP acseSAP, int proposedMaxPduSize, int proposedMaxServOutstandingCalling,
			int proposedMaxServOutstandingCalled, int proposedDataStructureNestingLevel,
			byte[] servicesSupportedCalling, int servicesSupportedCallingNumBits) throws IllegalArgumentException,
			ServiceError {

		negotiatedMaxPduSize = proposedMaxPduSize;

		associate(serverAccessPointReference, authenticationParameter, acseSAP, proposedMaxPduSize,
				proposedMaxServOutstandingCalling, proposedMaxServOutstandingCalled, proposedDataStructureNestingLevel,
				servicesSupportedCalling, servicesSupportedCallingNumBits);

		mmsScsmClientReceiver = new MmsScsmClientReceiver(acseAssociation, incomingResponses, incomingReports);
		mmsScsmClientReceiver.start();

	}

	private int getInvokeID() {
		invokeID = (invokeID + 1) % 2147483647;
		return invokeID;
	}

	private void readServerAPReference(String serverAccessPointReference) throws ServiceError, IllegalArgumentException {
		serverPort = 102;

		String[] serverAPRefSplits = serverAccessPointReference.split(":");
		if (serverAPRefSplits.length < 1 || serverAPRefSplits.length > 2) {
			throw new IllegalArgumentException("serverAccesPointReference argument is invalid");
		}
		if (serverAPRefSplits.length == 2) {
			try {
				serverPort = Integer.decode(serverAPRefSplits[1]);
			} catch (IllegalArgumentException e) {
				throw new IllegalArgumentException("serverAccesPointReference argument is invalid", e);
			}
			if (serverPort < 0 || serverPort > 65535) {
				throw new IllegalArgumentException("invalid port");
			}
		}

		try {
			serverAddress = InetAddress.getByName(serverAPRefSplits[0]);
		} catch (UnknownHostException e) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Hostname could not be resolved.", e);
		}
	}

	private static MmsPdu constructInitRequestPdu(int proposedMaxPduSize, int proposedMaxServOutstandingCalling,
			int proposedMaxServOutstandingCalled, int proposedDataStructureNestingLevel,
			byte[] servicesSupportedCalling, int servicesSupportedCallingNumBits) {

		InitRequestDetail initRequestDetail = new InitRequestDetail(version, proposedParameterCbbBitString,
				new BerBitString(servicesSupportedCalling, servicesSupportedCallingNumBits));

		InitiateRequestPdu initiateRequestPdu = new InitiateRequestPdu(new BerInteger(proposedMaxPduSize),
				new BerInteger(proposedMaxServOutstandingCalling), new BerInteger(proposedMaxServOutstandingCalled),
				new BerInteger(proposedDataStructureNestingLevel), initRequestDetail);

		MmsPdu initiateRequestMMSpdu = new MmsPdu(null, null, null, initiateRequestPdu, null, null);

		return initiateRequestMMSpdu;
	}

	private void handleInitiateResponse(MmsPdu responsePdu, int proposedMaxPduSize,
			int proposedMaxServOutstandingCalling, int proposedMaxServOutstandingCalled,
			int proposedDataStructureNestingLevel) throws ServiceError {

		testForErrorResponse(responsePdu);

		if (responsePdu.initiateResponsePdu == null) {
			acseAssociation.disconnect();
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding InitiateResponse Pdu");
		}

		InitiateResponsePdu initiateResponsePdu = responsePdu.initiateResponsePdu;

		if (initiateResponsePdu.localDetailCalled != null) {
			negotiatedMaxPduSize = (int) initiateResponsePdu.localDetailCalled.val;
		}

		int negotiatedMaxServOutstandingCalling = (int) initiateResponsePdu.negotiatedMaxServOutstandingCalling.val;
		int negotiatedMaxServOutstandingCalled = (int) initiateResponsePdu.negotiatedMaxServOutstandingCalled.val;

		int negotiatedDataStructureNestingLevel;
		if (initiateResponsePdu.negotiatedDataStructureNestingLevel != null) {
			negotiatedDataStructureNestingLevel = (int) initiateResponsePdu.negotiatedDataStructureNestingLevel.val;
		}
		else {
			negotiatedDataStructureNestingLevel = proposedDataStructureNestingLevel;
		}

		if (negotiatedMaxPduSize < MmsScsmClientSAP.MINIMUM_PDU_SIZE || negotiatedMaxPduSize > proposedMaxPduSize
				|| negotiatedMaxServOutstandingCalling > proposedMaxServOutstandingCalling
				|| negotiatedMaxServOutstandingCalling < 0
				|| negotiatedMaxServOutstandingCalled > proposedMaxServOutstandingCalled
				|| negotiatedMaxServOutstandingCalled < 0
				|| negotiatedDataStructureNestingLevel > proposedDataStructureNestingLevel
				|| negotiatedDataStructureNestingLevel < 0) {
			acseAssociation.disconnect();
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT, "Error negotiating paramters");
		}

		int version = (int) initiateResponsePdu.mmsInitResponseDetail.negotiatedVersionNumber.val;
		if (version != 1) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Unsupported version number was negotiated.");
		}

		byte[] servicesSupported = initiateResponsePdu.mmsInitResponseDetail.servicesSupportedCalled.bitString;
		if ((servicesSupported[0] & 0x40) != 0x40) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Obligatory services are not supported by the server.");
		}
	}

	private void associate(String serverAccessPointReference, String authenticationParameter, ClientAcseSAP acseSAP,
			int proposedMaxPduSize, int proposedMaxServOutstandingCalling, int proposedMaxServOutstandingCalled,
			int proposedDataStructureNestingLevel, byte[] servicesSupportedCalling, int servicesSupportedCallingNumBits)
			throws ServiceError, IllegalArgumentException {

		readServerAPReference(serverAccessPointReference);

		MmsPdu initiateRequestMMSpdu = constructInitRequestPdu(proposedMaxPduSize, proposedMaxServOutstandingCalling,
				proposedMaxServOutstandingCalled, proposedDataStructureNestingLevel, servicesSupportedCalling,
				servicesSupportedCallingNumBits);

		BerByteArrayOutputStream berOStream = new BerByteArrayOutputStream(500, true);
		try {
			initiateRequestMMSpdu.encode(berOStream, true);
		} catch (IOException e) {
			throw new ServiceError(ServiceError.FATAL, e);
		}

		try {
			acseAssociation = acseSAP.associate(serverAddress, serverPort, authenticationParameter,
					berOStream.getByteBuffer());

		} catch (IOException e) {
			if (acseAssociation != null) {
				acseAssociation.disconnect();
			}
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT, "Error starting association.",
					e);
		}
		ByteBuffer initResponse = acseAssociation.getAssociateResponseAPDU();

		MmsPdu initiateResponseMMSpdu = new MmsPdu();
		ByteArrayInputStream iStream = new ByteArrayInputStream(initResponse.array(), initResponse.arrayOffset()
				+ initResponse.position(), initResponse.limit() - initResponse.position());
		try {
			initiateResponseMMSpdu.decode(iStream, null);
		} catch (IOException e) {
			throw new ServiceError(ServiceError.FATAL, e);
		}

		handleInitiateResponse(initiateResponseMMSpdu, proposedMaxPduSize, proposedMaxServOutstandingCalling,
				proposedMaxServOutstandingCalled, proposedDataStructureNestingLevel);
	}

	/**
	 * The Model contains not SubDataObjects because they cannot be
	 * distinguished from Constructed Data Attributes in MMS. Also at the moment
	 * the model only contains FCDataObjects
	 */
	public Server retrieveModel() throws ServiceError {

		List<String> ldNames = retrieveLogicalDevices();
		List<List<String>> lnNames = new ArrayList<List<String>>(ldNames.size());

		for (int i = 0; i < ldNames.size(); i++) {
			lnNames.add(retrieveLogicalNodes(ldNames.get(i)));
		}
		List<LogicalDevice> lds = new ArrayList<LogicalDevice>();
		for (int i = 0; i < ldNames.size(); i++) {
			List<LogicalNode> lns = new ArrayList<LogicalNode>();
			for (int j = 0; j < lnNames.get(i).size(); j++) {
				lns.add(retrieveDataDefinitions(new ObjectReference(ldNames.get(i) + "/" + lnNames.get(i).get(j))));
			}
			lds.add(new LogicalDevice(new ObjectReference(ldNames.get(i)), lns));
		}

		retrieveDataSets(lds);

		model = new Server(lds);

		return model;
	}

	public ModelNode getDataValues(ModelNode modelNode) throws ServiceError {

		MmsPdu mmsReq = constructGetDataValuesRequest(modelNode);
		MmsPdu mmsRes = encodeWriteReadDecode(mmsReq);
		return decodeGetDataValuesResponse(mmsRes, modelNode);
	}

	private MmsPdu constructGetDataValuesRequest(ModelNode modelNode) {
		// Note: Array access based on IED Scout - there is also mention in MMS
		// documentation that AlternateAccess is used to identify the index
		// value of an array
		String objRef = modelNode.getReference().toString();
		AlternateAccess altAcc = checkArray(objRef);
		ObjectName objectName = null;

		if (altAcc == null) {
			objectName = MmsReference.toMmsObjectName(modelNode.getReference(), modelNode.getFunctionalConstraint());
		}

		else {
			objectName = MmsReference.toMmsObjectName(new ObjectReference(objRef.substring(0, objRef.indexOf('('))),
					modelNode.getFunctionalConstraint());
		}

		VariableSpecification varSpec = new VariableSpecification(objectName, null, null, null, null);

		ListOfVariableSeq seq = new ListOfVariableSeq(varSpec, altAcc);

		List<ListOfVariableSeq> listOfVariables = new ArrayList<ListOfVariableSeq>(1);
		listOfVariables.add(seq);

		VariableAccessSpecification varAccessSpec = new VariableAccessSpecification(new SubSeqOf_listOfVariable(
				listOfVariables), null);

		ReadRequest readRequest = new ReadRequest(null, varAccessSpec);

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(null, readRequest, null, null,
				null, null, null);

		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(getInvokeID()),
				confirmedServiceRequest);
		MmsPdu mms = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);
		return mms;
	}

	private ModelNode decodeGetDataValuesResponse(MmsPdu responsePdu, ModelNode modelNode) throws ServiceError {
		testForErrorResponse(responsePdu);

		if (responsePdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding GetDataValuesReponsePdu");
		}

		ConfirmedResponsePdu confirmedResponsePdu = responsePdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		if (confirmedServiceResponse.read == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding GetDataValuesReponsePdu");
		}

		ReadResponse readResponse = confirmedServiceResponse.read;
		List<AccessResult> listOfAccessResults = readResponse.listOfAccessResult.seqOf;

		if (listOfAccessResults.size() != 1) {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "Multiple results received.");
		}

		AccessResult accRes = listOfAccessResults.iterator().next();
		// first and only (?)
		if (accRes.failure == null) {
			MmsAsn1ComplexTypeConverter.convertMmsDataToDataContainer(modelNode, modelNode.getFunctionalConstraint(),
					new Data(accRes.array, accRes.structure, accRes.boolean_, accRes.bit_string, accRes.integer,
							accRes.unsigned, accRes.floating_point, accRes.octet_string, accRes.visible_string,
							accRes.generalized_time, accRes.binary_time, accRes.bcd, accRes.booleanArray,
							accRes.mMSString, accRes.utc_time), mmsFCs);

			return modelNode;
		}
		throw new ServiceError(mmsDataAccessErrorToServiceError(accRes.failure),
				"decodeGetDataValuesResponse: DataAccessError received.");
	}

	private int mmsDataAccessErrorToServiceError(BerInteger dataAccessError) {
		switch ((int) dataAccessError.val) {
		case 1: // hardware_fault
			return ServiceError.FAILED_DUE_TO_SERVER_CONSTRAINT;
		case 2: // temporarily-unavailable
			return ServiceError.INSTANCE_LOCKED_BY_OTHER_CLIENT;
		case 3: // object-access-denied
			return ServiceError.ACCESS_VIOLATION;
		case 5: // invalid-address
			return ServiceError.PARAMETER_VALUE_INCONSISTENT;
		case 7: // type-inconsistent
			return ServiceError.TYPE_CONFLICT;
		case 10: // object-non-existent
			return ServiceError.INSTANCE_NOT_AVAILABLE;
		default:
			return ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT; // ?
		}
	}

	private MmsPdu constructGetLDDirectoryRequest(String ldRef, String continueAfter, boolean logicalDevice) {
		// TODO speed this up by preparing the mmspdu in the constructor

		ObjectClass objectClass = null;

		if (logicalDevice) {
			objectClass = new ObjectClass(new BerInteger(MmsConstants.NAMED_VARIABLE));
		}
		else { // for data sets
			objectClass = new ObjectClass(new BerInteger(MmsConstants.NAMED_VARIABLE_LIST));
		}

		GetNameListRequest getNameListRequest = null;
		try {
			SubChoice_objectScope objectScopeChoiceType = new SubChoice_objectScope(null, new BerVisibleString(ldRef),
					null);

			if (continueAfter != "") {
				getNameListRequest = new GetNameListRequest(objectClass, objectScopeChoiceType, new BerVisibleString(
						continueAfter));
			}
			else {
				getNameListRequest = new GetNameListRequest(objectClass, objectScopeChoiceType, null);
			}
		} catch (UnsupportedEncodingException e) {
			throw new IllegalAccessError();
		}

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(getNameListRequest, null, null,
				null, null, null, null);
		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(getInvokeID()),
				confirmedServiceRequest);
		MmsPdu confirmedRequestMMSPdu = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);

		return confirmedRequestMMSPdu;
	}

	private MmsPdu constructGetDataDefinitionRequest(ObjectReference lnRef) {

		SubSeq_domain_specific domainSpec = null;
		try {
			domainSpec = new SubSeq_domain_specific(new BerVisibleString(lnRef.get(0)), new BerVisibleString(
					lnRef.get(1)));
		} catch (UnsupportedEncodingException e) {
			throw new IllegalAccessError();
		}

		ObjectName objectName = new ObjectName(null, domainSpec, null);

		GetVariableAccessAttributesRequest getVariableAccessAttributesRequest = new GetVariableAccessAttributesRequest(
				objectName, null);
		// getVariableAccessAttributesRequest.selectName(objectName);

		// GetNameListRequest getNameListRequest = new GetNameListRequest();
		// getNameListRequest.setObjectClass(objectClass);
		// DomainSpecificChoiceType domainSpecificChoiceType = new
		// DomainSpecificChoiceType();
		// domainSpecificChoiceType.selectBasic(new BasicIdentifier(new
		// String(ldRef.getValue())));
		// ObjectScopeChoiceType objectScopeChoiceType = new
		// ObjectScopeChoiceType();
		// //
		// objectScopeChoiceType.selectDomainSpecific(domainSpecificChoiceType);
		// objectScopeChoiceType.selectDomainSpecific(new
		// Identifier(ldRef.getValue()));
		// getNameListRequest.setObjectScope(objectScopeChoiceType);

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(null, null, null,
				getVariableAccessAttributesRequest, null, null, null);
		// confirmedServiceRequest.selectGetNameList(getNameListRequest);
		// confirmedServiceRequest
		// .selectGetVariableAccessAttributes(getVariableAccessAttributesRequest);

		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(getInvokeID()),
				confirmedServiceRequest);
		// confirmedRequestPdu.setInvokeID(new Unsigned32(Long
		// .valueOf(getInvokeID())));
		// confirmedRequestPdu.setConfirmedServiceRequest(confirmedServiceRequest);
		MmsPdu confirmedRequestMMSPdu = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);
		// confirmedRequestMMSPdu.selectConfirmedRequestPdu(confirmedRequestPdu);

		// ContinueAfterChoiceType continueAfterCT = new
		// ContinueAfterChoiceType();
		// continueAfterCT.selectBasic(new BasicIdentifier(continueAfter));
		// if (continueAfter != "") {
		// // getNameListRequest.setContinueAfter(continueAfterCT);
		// getNameListRequest.setContinueAfter(new Identifier(continueAfter));
		// }

		return confirmedRequestMMSPdu;
	}

	//
	// public List<String> getLogicalDeviceDirectory(String ldRef) throws
	// ServiceError {
	// LogicalDevice ld = findOrRetrieveLogicalDevice(ldRef);
	// if (ld == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE);
	// }
	//
	// retrieveLogicalNodes(ld);
	//
	// ArrayList<String> lnReferences = new ArrayList<String>();
	// for (LogicalNode ln : ld.getLNodes()) {
	// lnReferences.add(ln.getReference().toString());
	// }
	// return lnReferences;
	// }
	//
	public List<String> retrieveLogicalNodes(String ld) throws ServiceError {
		List<String> lns = new LinkedList<String>();
		String continueAfterRef = "";
		do {
			MmsPdu getLDDirectoryRequestMMSPdu = constructGetLDDirectoryRequest(ld, continueAfterRef, true);
			MmsPdu getLDDirectoryResponseMMSPdu = encodeWriteReadDecode(getLDDirectoryRequestMMSPdu);
			continueAfterRef = decodeGetDirectoryResponse(getLDDirectoryResponseMMSPdu, lns);

		} while (continueAfterRef != "");
		return lns;
	}

	/**
	 * Decodes an MMS response which contains the structure of a LD and its LNs
	 * including names of DOs.
	 */
	private String decodeGetDirectoryResponse(MmsPdu getLDDirectoryResponseMmsPdu, List<String> lns)
			throws ServiceError {

		testForErrorResponse(getLDDirectoryResponseMmsPdu);

		// if (!getLDDirectoryResponseMmsPdu.isConfirmedResponsePduSelected()) {
		if (getLDDirectoryResponseMmsPdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeGetLDDirectoryResponse: Error decoding server response");
		}

		ConfirmedResponsePdu confirmedResponsePdu = getLDDirectoryResponseMmsPdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		// if (!confirmedServiceResponse.isGetNameListSelected()) {
		if (confirmedServiceResponse.getNameList == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeGetLDDirectoryResponse: Error decoding server response");
		}

		GetNameListResponse getNameListResponse = confirmedServiceResponse.getNameList;

		List<BerVisibleString> identifiers = getNameListResponse.listOfIdentifier.seqOf;

		if (identifiers.size() == 0) {
			throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
					"decodeGetLDDirectoryResponse: Instance not available");
		}

		BerVisibleString identifier = null;
		Iterator<BerVisibleString> it = identifiers.iterator();

		String idString;

		while (it.hasNext()) {
			identifier = it.next();
			idString = identifier.toString();

			if (idString.indexOf('$') == -1) {
				lns.add(idString);
			}
		}

		if (getNameListResponse.moreFollows != null && getNameListResponse.moreFollows.val == false) {
			return "";
		}
		else {
			return identifier.toString();
		}
	}

	public static void testForErrorResponse(MmsPdu mmsResponsePdu) throws ServiceError {
		if (mmsResponsePdu.initiateErrorPdu == null) {
			return;
		}

		// InitiateErrorPdu errorPdu = mmsResponsePdu.initiateErrorPdu;
		org.openiec61850.jmms.mms.asn1.ServiceError asn1ServiceErr = mmsResponsePdu.initiateErrorPdu;

		// ErrorClassChoiceType errClass = asn1ServiceErr.getErrorClass();
		SubChoice_errorClass errClass = asn1ServiceErr.errorClass;
		if (errClass.others != null) { // TODO auf andere Fehlerklassen
			// prüfen
			throw new ServiceError((int) errClass.others.val); // TODO
			// geraten
		}
		throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT);
	}

	private static MmsPdu constructGetServerDirectoryRequest(int invokeID) {
		ObjectClass objectClass = new ObjectClass(new BerInteger(MmsConstants.DOMAIN));
		// objectClass.selectBasicObjectClass((long) MMSConstants.DOMAIN);
		// the previous statement was replace the following to imitate a siemens
		// iec client:
		// MMSConstants.NAMED_VARIABLE);
		// MMSConstants.NAMED_VARIABLE_LIST);

		// GetNameListRequest getNameListRequest = new GetNameListRequest();
		// getNameListRequest.setObjectClass(objectClass);

		GetNameListRequest.SubChoice_objectScope objectScope = new GetNameListRequest.SubChoice_objectScope(
				new BerNull(), null, null);
		// ObjectScopeChoiceType objectScopeChoiceType = new
		// ObjectScopeChoiceType();
		// objectScopeChoiceType.selectVmdSpecific();
		// objectScopeChoiceType.selectAaSpecific();

		GetNameListRequest getNameListRequest = new GetNameListRequest(objectClass, objectScope, null);

		// getNameListRequest.setObjectScope(objectScopeChoiceType);

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(getNameListRequest, null, null,
				null, null, null, null);

		// ConfirmedServiceRequest confirmedServiceRequest = new
		// ConfirmedServiceRequest();
		// confirmedServiceRequest.selectGetNameList(getNameListRequest);

		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(invokeID),
				confirmedServiceRequest);
		// ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu();
		// confirmedRequestPdu.setInvokeID(new
		// Unsigned32(Long.valueOf(getInvokeID())));
		// confirmedRequestPdu.setConfirmedServiceRequest(confirmedServiceRequest);

		MmsPdu confirmedRequestMMSPdu = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);
		// MmsPdu confirmedRequestMMSPdu = new MmsPdu();
		// confirmedRequestMMSPdu.selectConfirmedRequestPdu(confirmedRequestPdu);

		return confirmedRequestMMSPdu;

	}

	public static void checkInvokeID(ConfirmedResponsePdu confirmedResponsePdu, int invokeID) throws ServiceError {
		if (confirmedResponsePdu.invokeID.val != invokeID) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding Get Server Directory Response Pdu: incorrect invokeID");
		}
	}

	private List<String> decodeGetServerDirectoryResponse(MmsPdu getServerDirectoryResponseMmsPdu) throws ServiceError {
		testForErrorResponse(getServerDirectoryResponseMmsPdu);

		if (getServerDirectoryResponseMmsPdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding Get Server Directory Response Pdu");
		}

		ConfirmedResponsePdu confirmedResponsePdu = getServerDirectoryResponseMmsPdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		if (confirmedServiceResponse.getNameList == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding Get Server Directory Response Pdu");
		}

		List<BerVisibleString> identifiers = confirmedServiceResponse.getNameList.listOfIdentifier.seqOf;
		ArrayList<String> objectRefs = new ArrayList<String>(); // ObjectReference[identifiers.size()];

		for (BerVisibleString identifier : identifiers) {
			objectRefs.add(identifier.toString());
		}

		return objectRefs;
	}

	private List<String> retrieveLogicalDevices() throws ServiceError {
		MmsPdu requestPdu = constructGetServerDirectoryRequest(getInvokeID());
		MmsPdu responsePdu = encodeWriteReadDecode(requestPdu);
		return decodeGetServerDirectoryResponse(responsePdu);
	}

	/**
	 * "Encode, write, read, decode"
	 */
	private MmsPdu encodeWriteReadDecode(MmsPdu requestPdu) throws ServiceError {
		BerByteArrayOutputStream baos = new BerByteArrayOutputStream(500, true);
		try {
			// encoder.encode(requestPdu, baos);
			requestPdu.encode(baos, true);
		} catch (Exception e) {
			acseAssociation.disconnect();
			throw new RuntimeException("Error encoding MmsPdu.", e);
		}

		mmsScsmClientReceiver.setResponseExpected();
		try {
			acseAssociation.send(baos.getByteBuffer());
		} catch (IOException e) {
			acseAssociation.disconnect();
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT, "Error sending packet.", e);
		}

		// ByteBuffer encodedResponsePdu;
		// try {
		// encodedResponsePdu = acseAssociation.receive();
		// } catch (IOException e) {
		// acseAssociation.disconnect();
		// throw new ServiceError(
		// ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
		// "Error receiving packet.", e);
		// }
		//
		// MmsPdu decodedResponsePdu = new MmsPdu();
		// try {
		// decodedResponsePdu.decode(new ByteArrayInputStream(
		// encodedResponsePdu.array(), encodedResponsePdu
		// .arrayOffset()
		// + encodedResponsePdu.position(), encodedResponsePdu
		// .limit()
		// - encodedResponsePdu.position()), null);
		//
		// } catch (Exception e) {
		// throw new ServiceError(
		// ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
		// "Error decoding MmsPdu", e);
		// }

		MmsPdu decodedResponsePdu = null;
		try {
			while (decodedResponsePdu == null && mmsScsmClientReceiver.isAlive()) {
					// Actually, we would need yet another timout setting here, a timeout for receiving a
					// complete message (initial timeout + all fragment timeouts). Since we do not know how
					// many fragments will be used, we re-use the message timeout and loop if there is a chance
					// of getting more data. Without any timeout (when using take() on the queue), this thread
					// would be stuck if the client reciever thread dies.
				decodedResponsePdu = incomingResponses.poll(acseAssociation.getMessageTimeout(), TimeUnit.MILLISECONDS);
			}
		} catch (InterruptedException e) {
		}
		return decodedResponsePdu;
	}

	protected LogicalNode retrieveDataDefinitions(ObjectReference lnRef) throws ServiceError {
		MmsPdu getDataDefinitionRequestMMSPdu = constructGetDataDefinitionRequest(lnRef);
		MmsPdu getDataDefinitionResponseMMSPdu = encodeWriteReadDecode(getDataDefinitionRequestMMSPdu);
		return DataDefinitionResParser.parseGetDataDefinitionResponse(getDataDefinitionResponseMMSPdu, lnRef, invokeID);
	}

	public void setDataValues(ModelNode container) throws ServiceError {
		MmsPdu mmsReq = constructSetDataValuesRequest(container);
		MmsPdu mmsRes = encodeWriteReadDecode(mmsReq);
		handleSetDataValuesResponse(mmsRes);
	}

	private MmsPdu constructSetDataValuesRequest(ModelNode container) throws ServiceError {
		// Create and configure Pdus in reverse order
		String objRef = container.getReference().toString();
		AlternateAccess altAcc = checkArray(objRef);
		ObjectName objectName = null;

		if (altAcc == null) {
			objectName = MmsReference.toMmsObjectName(container.getReference(), container.getFunctionalConstraint());
		}
		else {
			objectName = MmsReference.toMmsObjectName(new ObjectReference(objRef.substring(0, objRef.indexOf('('))),
					container.getFunctionalConstraint());
		}

		VariableSpecification varSpec = new VariableSpecification(objectName, null, null, null, null);

		List<ListOfVariableSeq> listOfVariable = new LinkedList<ListOfVariableSeq>();
		ListOfVariableSeq varSeq = new ListOfVariableSeq(varSpec, altAcc);
		listOfVariable.add(varSeq);

		VariableAccessSpecification variableAccessSpec = new VariableAccessSpecification(new SubSeqOf_listOfVariable(
				listOfVariable), null);

		List<Data> listOfData = new LinkedList<Data>();
		Data mmsData = MmsAsn1ComplexTypeConverter.convertModelNodeToMmsData(container);
		listOfData.add(mmsData);

		WriteRequest writeRequest = new WriteRequest(variableAccessSpec, new SubSeqOf_listOfData(listOfData));

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(null, null, writeRequest, null,
				null, null, null);

		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(getInvokeID()),
				confirmedServiceRequest);

		MmsPdu mms = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);

		return mms;
	}

	public void retrieveDataSets(List<LogicalDevice> lds) throws ServiceError {
		for (ModelNode ld : lds) {
			String continueAfterRef = "";
			do {
				MmsPdu getDSDriectoryRequestMMSPdu = constructGetLDDirectoryRequest(ld.getNodeName(), continueAfterRef,
						false);
				MmsPdu getLDDirectoryResponseMMSPdu = encodeWriteReadDecode(getDSDriectoryRequestMMSPdu);
				continueAfterRef = decodeGetDataSetResponse(getLDDirectoryResponseMMSPdu, (LogicalDevice) ld);

			} while (continueAfterRef != "");
		}

	}

	private String decodeGetDataSetResponse(MmsPdu getLDDirectoryResponseMMSPdu, LogicalDevice ld) throws ServiceError {

		testForErrorResponse(getLDDirectoryResponseMMSPdu);

		if (getLDDirectoryResponseMMSPdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeGetDataSetResponse: Error decoding server response");
		}

		ConfirmedResponsePdu confirmedResponsePdu = getLDDirectoryResponseMMSPdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		if (confirmedServiceResponse.getNameList == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeGetDataSetResponse: Error decoding server response");
		}

		GetNameListResponse getNameListResponse = confirmedServiceResponse.getNameList;

		List<BerVisibleString> identifiers = getNameListResponse.listOfIdentifier.seqOf;

		if (identifiers.size() == 0) {
			// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
			// "decodeGetDataSetResponse: Instance not available");
			logger.info("There are no data sets");
			return "";
		}

		String identifier = null;
		Iterator<BerVisibleString> it = identifiers.iterator();

		while (it.hasNext()) {
			identifier = ld.getNodeName() + "/" + it.next().toString();

			// response is always of type LN$DS_Name
			if (identifier.indexOf('$') == -1) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
						"decodeGetDataSetResponse: Error decoding server response");
			}

			identifier = identifier.replace("$", ".");

			getDataSetDirectory(identifier, ld);
		}

		if (getNameListResponse.moreFollows != null && getNameListResponse.moreFollows.val == false) {
			return "";
		}
		else {
			return identifier.toString();
		}
	}

	private void getDataSetDirectory(String dataSetRef, LogicalDevice ld) throws ServiceError {
		MmsPdu mmsReq = constructGetDataSetDirectory(dataSetRef);
		MmsPdu mmsRes = encodeWriteReadDecode(mmsReq);
		decodeGetDataSetDirectoryResponse(mmsRes, dataSetRef, ld);
	}

	// /*
	// * private static Data convertModelNodeToMmsData(IModelNode node) throws
	// * ServiceError { switch(node.getObjectType()) { case SERVER: //not
	// allowed,
	// * but possible case LOGICAL_DEVICE: //not allowed, but possible case
	// * LOGICAL_NODE: //not allowed, but possible case DATA_OBJECT: case
	// * CONSTRUCTED_DATA_ATTRIBUTE: case DATA_SET: Collection<Data> listOfData
	// =
	// * new LinkedList<Data>(); for(IModelNode child : node) {
	// * listOfData.add(convertModelNodeToMmsData(child)); } Data mmsStruct =
	// new
	// * Data(); mmsStruct.selectStructure(new DataSequence(listOfData)); return
	// * mmsStruct;
	// *
	// * case ARRAY_OF_DO: case ARRAY_OF_DA: Collection<Data> arrayOfData = new
	// * LinkedList<Data>(); for(IModelNode child : node) {
	// * arrayOfData.add(convertModelNodeToMmsData(child)); } Data mmsArray =
	// new
	// * Data(); mmsArray.selectArray(new DataSequence(arrayOfData)); return
	// * mmsArray;
	// *
	// * case BASIC_TYPE: return convertBasicNodeToMmsData(node);
	// *
	// * default: throw new
	// * ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
	// * "SetDataValues: unsupported ObjectType"); } }
	// *
	// * private static Data convertBasicNodeToMmsData(IModelNode node) {
	// * switch(node.getBasicType()) { //case } return null; }
	// */
	//
	private void handleSetDataValuesResponse(MmsPdu mmsRes) throws ServiceError {
		testForErrorResponse(mmsRes);

		// there was no service error but might still be a write failure
		if (mmsRes.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"handleSetDataValuesResponse: improper response");
		}

		ConfirmedResponsePdu confRes = mmsRes.confirmedResponsePdu;

		if (confRes.confirmedServiceResponse == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"handleSetDataValuesResponse: improper response");
		}

		ConfirmedServiceResponse confSerRes = confRes.confirmedServiceResponse;

		if (confSerRes.write == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"handleSetDataValuesResponse: improper response");
		}

		WriteResponse response = confSerRes.write;

		// only one?
		WriteResponse.SubChoice subChoice = response.seqOf.get(0);

		if (subChoice.failure != null) {
			throw new ServiceError(ServiceError.DATA_ACCESS_ERROR);
		}
	}

	// Note: fc and dsMemberRef should correspond one to one
	// TODO the input can be a list of fcDataObjects nodes instead of the names
	// and functional constraints
	// public void createDataSet(String dsRef, List<String> dsMemberRef,
	// List<FunctionalConstraint> fc) throws ServiceError {
	// store the data set on client side
	// int at = dsRef.indexOf('@');
	// ModelNode mem;
	// FunctionalConstraint funcCon = null;
	// Iterator<String> itMem = dsMemberRef.iterator();
	// Iterator<FunctionalConstraint> itFC = fc.iterator();
	// String member;
	// LogicalNode lnNode = null;

	// if (dsMemberRef.size() != fc.size()) {
	// throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
	// "createDataSet: the functional constraints do not match the list of members");
	// }
	//
	// DataSet dsTest = retrieveDataSet(dsRef);
	// if (dsTest != null) {
	// throw new ServiceError(ServiceError.INSTANCE_IN_USE,
	// "createDataSet: already have data set with same name");
	// }
	//
	// DataSet dataSet = new DataSet(dsRef);
	//
	// // get the members
	// while (itMem.hasNext()) {
	// funcCon = itFC.next();
	// member = itMem.next();
	// mem = findOrRetrieveNode(member);
	// if (mem == null) {
	// mem = getDataDefinition(member);
	// }
	// if (mem.getObjectType() == ObjectType.DATA_OBJECT) {
	// DataObject obj = new DataObject(mem.getReference().toString());
	// getFuncConstrainedDataObject(mem, funcCon, obj);
	// mem = obj;
	// }
	// dataSet.addChild((DataContainer) mem);
	// }
	// at this point the data set is deletable
	// dataSet.setDeletable(true);

	// if (at < 0) {
	// int dot = dsRef.indexOf('.');
	// lnNode = (LogicalNode) model.findNode(dsRef.substring(0, dot));
	// if (lnNode == null) {// getLogicalDeviceDirectory not called
	// throw new ServiceError(
	// ServiceError.PARAMETER_VALUE_INCONSISTENT,
	// "Logical node not in model: call getLogicalDeviceDirectory");
	// }
	// lnNode.addDataSet(dataSet);
	// } else {
	// nonPersistentDS.add(dataSet);
	// }

	/**
	 * The client should create the data set first and add it to either the
	 * nonpersistent list or to the model. Then it should call this method for
	 * creation on the server side
	 */
	public void createDataSet(DataSet dataSet) throws ServiceError {
		// send mms msg to server to create the data set
		MmsPdu mmsReq = constructCreateDataSetRequest(dataSet);
		MmsPdu mmsRes = encodeWriteReadDecode(mmsReq);
		handleCreateDataSetResponse(mmsRes);
	}

	/*
	 * dsRef = either LD/LN.DataSetName (persistent) or @DataSetname
	 * (non-persistent) Names in dsMemberRef should be in the form:
	 * LD/LNName.DoName or LD/LNName.DoName.DaName
	 */
	private MmsPdu constructCreateDataSetRequest(DataSet dataSet) throws ServiceError {
		// construct object name
		ObjectName dataSetObjName = MmsReference.toMmsObjectNameDS(dataSet.getReference().toString());

		// set the list of variables
		List<DefineNamedVariableListRequest.SubSeqOf_listOfVariable.SubSeq> listOfVariable = new LinkedList<DefineNamedVariableListRequest.SubSeqOf_listOfVariable.SubSeq>();
		FunctionalConstraint fc = null;

		for (ModelNode dc : dataSet.getChildren()) {
			fc = dc.getFunctionalConstraint();

			ObjectName dsMember = MmsReference.toMmsObjectName(dc.getReference(), fc);
			VariableSpecification memVarSpec = new VariableSpecification(dsMember, null, null, null, null);
			listOfVariable.add(new DefineNamedVariableListRequest.SubSeqOf_listOfVariable.SubSeq(memVarSpec, null));
		}

		DefineNamedVariableListRequest.SubSeqOf_listOfVariable seqOf = new DefineNamedVariableListRequest.SubSeqOf_listOfVariable(
				listOfVariable);
		DefineNamedVariableListRequest createDSRequest = new DefineNamedVariableListRequest(dataSetObjName, seqOf);

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(null, null, null, null,
				createDSRequest, null, null);

		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(invokeID),
				confirmedServiceRequest);

		MmsPdu mms = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);

		return mms;

	}

	private void handleCreateDataSetResponse(MmsPdu responsePdu) throws ServiceError {
		// check for errors
		testForErrorResponse(responsePdu);
	}

	private MmsPdu constructGetDataSetDirectory(String dsRef) throws ServiceError {
		ObjectName dataSetObj = MmsReference.toMmsObjectNameDS(dsRef);
		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(null, null, null, null, null,
				dataSetObj, null);
		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(invokeID),
				confirmedServiceRequest);
		MmsPdu mms = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);
		return mms;
	}

	private void decodeGetDataSetDirectoryResponse(MmsPdu responsePdu, String dsRef, LogicalDevice ld)
			throws ServiceError {
		// check response msg
		testForErrorResponse(responsePdu);

		if (responsePdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeGetDataSetDirectoryResponse: Error decoding server response");
		}

		ConfirmedResponsePdu confirmedResponsePdu = responsePdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		if (confirmedServiceResponse.getNamedVariableListAttributes == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeGetDataSetDirectoryResponse: Error decoding server response");
		}

		GetNamedVariableListAttributesResponse getNamedVariableListAttResponse = confirmedServiceResponse.getNamedVariableListAttributes;
		boolean deletable = getNamedVariableListAttResponse.mmsDeletable.val;
		List<GetNamedVariableListAttributesResponse.SubSeqOf_listOfVariable.SubSeq> variables = getNamedVariableListAttResponse.listOfVariable.seqOf;

		if (variables.size() == 0) {
			throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
					"decodeGetDataSetDirectoryResponse: Instance not available");
		}

		List<ModelNode> dsMems = new ArrayList<ModelNode>();

		// add members to data set
		VariableSpecification memVS;
		ObjectReference memberRef;
		FunctionalConstraint fc;

		for (GetNamedVariableListAttributesResponse.SubSeqOf_listOfVariable.SubSeq subSeq : variables) {
			memVS = subSeq.variableSpecification;
			memberRef = MmsReference.toObjectReference(memVS.name.domain_specific.domainId.toString(),
					memVS.name.domain_specific.itemId.toString());
			fc = MmsReference.getFunctionalConstraint(memVS.name.domain_specific.itemId.toString());

			ModelNode member = ld.findSubNode(memberRef, fc);
			if (member == null) {
				throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
						"decodeGetDataSetDirectoryResponse: data set memeber does not exist, call getDataDefinition first");
			}
			dsMems.add(member);
		}

		ObjectReference dsObjRef = new ObjectReference(dsRef);
		DataSet dataSet = new DataSet(dsObjRef, dsMems, deletable);
		ModelNode ln = ld.findSubNode(new ObjectReference(dsObjRef.getParentRef()), null);
		if (ln == null) {
			throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
					"decodeGetDataSetDirectoryResponse: Instance not available");
		}

		((LogicalNode) ln).addDataSet(dataSet);

	}

	// // int[] has format: int[0] = numberMatched, int[1] = numberDeleted
	// The client should delete the data set from the logical
	// node/non-persistent list and then call
	// this method to have it deleted from the server

	public int[] deleteDataSet(DataSet dataSet) throws ServiceError {
		// first find dataSet in model
		// int at = dataSetRef.indexOf('@');
		// DataSet dataSet = null;
		//
		// if (at < 0) {
		// int dot = dataSetRef.indexOf('.');
		// LogicalNode lnNode = (LogicalNode) findOrRetrieveNode(dataSetRef
		// .substring(0, dot));
		// if (lnNode == null) {
		// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
		// "cannot find logical node");
		// }
		// dataSet = lnNode.getDataSet(dataSetRef.substring(dot + 1));
		// if (dataSet == null || !dataSet.isDeletable()) {
		// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
		// "Data set does not exist or cannot be deleted");
		// }
		// lnNode.deleteDataSet(dataSet.getNodeName());
		// }
		//
		// else {
		// for (DataSet ds : nonPersistentDS) {
		// if (dataSetRef.equals(ds.getReference().toString())) {
		// if (!ds.isDeletable()) {
		// throw new ServiceError(
		// ServiceError.INSTANCE_NOT_AVAILABLE,
		// "Data set does not exist or cannot be deleted");
		// }
		// dataSet = ds;
		// nonPersistentDS.remove(ds);
		// break;
		// }
		// }
		//
		// if (dataSet == null) {
		// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
		// "Data set does not exist or cannot be deleted");
		// }
		// }

		// delete the data set on the server side
		MmsPdu request = constructDeleteDataSetRequest(dataSet);
		MmsPdu response = encodeWriteReadDecode(request);
		return decodeDeleteDataSet(response);
	}

	private MmsPdu constructDeleteDataSetRequest(DataSet dataSet) throws ServiceError {
		ObjectName objName = MmsReference.toMmsObjectNameDS(dataSet.getReference().toString());
		List<ObjectName> listOfVariableListName = new LinkedList<ObjectName>();
		// only deleting one element
		listOfVariableListName.add(objName);

		DeleteNamedVariableListRequest requestDeleteDS = new DeleteNamedVariableListRequest(null,
				new SubSeqOf_listOfVariableListName(listOfVariableListName), null);

		ConfirmedServiceRequest requestService = new ConfirmedServiceRequest(null, null, null, null, null, null,
				requestDeleteDS);

		ConfirmedRequestPdu request = new ConfirmedRequestPdu(new BerInteger(invokeID), requestService);

		MmsPdu mmsRequest = new MmsPdu(request, null, null, null, null, null);

		return mmsRequest;
	}

	private int[] decodeDeleteDataSet(MmsPdu responsePdu) throws ServiceError {
		testForErrorResponse(responsePdu);

		if (responsePdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeDeleteDataSetResponse: Error decoding server response");
		}

		ConfirmedResponsePdu confirmedResponsePdu = responsePdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		if (confirmedServiceResponse.deleteNamedVariableList == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"decodeDeleteDataSetResponse: Error decoding server response");
		}

		DeleteNamedVariableListResponse deleteNamedVariableListResponse = confirmedServiceResponse.deleteNamedVariableList;

		int[] numMatchedDeleted = new int[2];
		numMatchedDeleted[0] = (int) deleteNamedVariableListResponse.numberMatched.val;
		numMatchedDeleted[1] = (int) deleteNamedVariableListResponse.numberDeleted.val;

		return numMatchedDeleted;

	}

	public void getDataSetValues(DataSet dataSet) throws ServiceError {
		// always check server for updated values
		// get the data set since only the values are read

		if (dataSet.getChildren() == null) {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE,
					"getDataSetValues: data set has no children, call getDataSetDirectory first!!");
		}

		MmsPdu request = constructGetDataSetValuesRequest(dataSet);
		MmsPdu responsePdu = encodeWriteReadDecode(request);
		decodeGetDataSetValuesResponse(responsePdu, dataSet);
	}

	private MmsPdu constructGetDataSetValuesRequest(DataSet dataSet) throws ServiceError {
		// TODO merge with constructGetDataValuesRequest??
		ObjectName dsObjName = MmsReference.toMmsObjectNameDS(dataSet.getReference().toString());
		VariableAccessSpecification varAccSpec = new VariableAccessSpecification(null, dsObjName);
		ReadRequest getDataSetValuesRequest = new ReadRequest(new BerBoolean(true), varAccSpec);

		// getDataSetValuesRequest.setSpecificationWithResult(true);
		// left as default of 'false' for getDataValues

		// getDataSetValuesRequest.setVariableAccessSpecification(varAccSpec);

		ConfirmedServiceRequest requestService = new ConfirmedServiceRequest(null, getDataSetValuesRequest, null, null,
				null, null, null);
		ConfirmedRequestPdu request = new ConfirmedRequestPdu(new BerInteger(invokeID), requestService);
		MmsPdu mmsRequest = new MmsPdu(request, null, null, null, null, null);

		return mmsRequest;

	}

	private void decodeGetDataSetValuesResponse(MmsPdu responsePdu, DataSet ds) throws ServiceError {
		// check response for error
		testForErrorResponse(responsePdu);

		if (responsePdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding GetDataValuesReponsePdu");
		}

		ConfirmedResponsePdu confirmedResponsePdu = responsePdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		if (confirmedServiceResponse.read == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding GetDataValuesReponsePdu");
		}

		ReadResponse readResponse = confirmedServiceResponse.read;
		List<AccessResult> listOfAccessResults = readResponse.listOfAccessResult.seqOf;

		if (listOfAccessResults.size() != ds.getChildren().size()) {
			// should be one AccessResult for each data set memeber
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "Too many results received.");
		}

		Iterator<AccessResult> it = listOfAccessResults.iterator();
		AccessResult accRes = null;

		Iterator<ModelNode> dsIt = ds.getChildren().iterator();
		ModelNode member = null;

		// List<ModelNode> results = new LinkedList<ModelNode>();
		int failures = 0;
		while (it.hasNext()) {
			accRes = it.next();
			member = dsIt.next();
			if (accRes.failure == null) {
				// if (member instanceof DataObject) {
				// for (DataContainer child : member.recursive()) {
				// if (child instanceof ConstructedDataAttribute
				// && child.getFunctionalConstraint() != null) {
				// fc = child.getFunctionalConstraint();
				// break;
				// }
				// }

				MmsAsn1ComplexTypeConverter.convertMmsDataToDataContainer(member, member.getFunctionalConstraint(),
						new Data(accRes.array, accRes.structure, accRes.boolean_, accRes.bit_string, accRes.integer,
								accRes.unsigned, accRes.floating_point, accRes.octet_string, accRes.visible_string,
								accRes.generalized_time, accRes.binary_time, accRes.bcd, accRes.booleanArray,
								accRes.mMSString, accRes.utc_time), mmsFCs);
				// results.add(member);

				// } else {
				// fc = member.getFunctionalConstraint();
				// results.add(MmsAsn1ComplexTypeConverter
				// .convertMmsDataToDataContainer(member, fc, accRes
				// .getSuccess()));
				// }

			}
			else {
				failures++;
				// results.add(null);
				// leave to client to handle?
			}
		}

		// indicate failure this way?
		if (failures != 0) {
			logger.error("decodeGetDataSetValuesResponse: " + failures + " DataAccessError received.");
		}

		// return results;

	}

	/**
	 * The client modifies the values of the data set first then passes the data
	 * set to this function for changes on the server side
	 */
	public List<String> setDataSetValues(DataSet dataSet) throws ServiceError {

		// prepare, send msg and decode response
		MmsPdu mmsRequest = constructSetDataSetValues(dataSet);
		MmsPdu mmsResponse = encodeWriteReadDecode(mmsRequest);
		return decodeSetDataSetValues(mmsResponse);

	}

	private MmsPdu constructSetDataSetValues(DataSet dataSet) throws ServiceError {
		// insert data set name
		ObjectName dsName = MmsReference.toMmsObjectNameDS(dataSet.getReference().toString());
		VariableAccessSpecification varAccSpec = new VariableAccessSpecification(null, dsName);

		// construct list of data
		List<Data> listOfData = new LinkedList<Data>();

		for (ModelNode member : dataSet.getChildren()) {
			listOfData.add(MmsAsn1ComplexTypeConverter.convertModelNodeToMmsData(member));
		}

		WriteRequest setDataSetValuesRequest = new WriteRequest(varAccSpec, new SubSeqOf_listOfData(listOfData));
		ConfirmedServiceRequest requestService = new ConfirmedServiceRequest(null, null, setDataSetValuesRequest, null,
				null, null, null);
		ConfirmedRequestPdu request = new ConfirmedRequestPdu(new BerInteger(invokeID), requestService);
		MmsPdu mmsRequest = new MmsPdu(request, null, null, null, null, null);
		return mmsRequest;
	}

	private List<String> decodeSetDataSetValues(MmsPdu responsePdu) throws ServiceError {
		testForErrorResponse(responsePdu);

		if (responsePdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding SetDataSetValuesReponsePdu");
		}

		ConfirmedResponsePdu confirmedResponsePdu = responsePdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		if (confirmedServiceResponse.write == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding SetDataSetValuesReponsePdu");
		}

		WriteResponse writeResponse = confirmedServiceResponse.write;
		List<WriteResponse.SubChoice> writeResChoiceType = writeResponse.seqOf;
		List<String> response = new ArrayList<String>(writeResChoiceType.size());

		for (WriteResponse.SubChoice result : writeResChoiceType) {

			if (result.failure != null) {
				response.add("Failure: reason = " + result.failure.val);
			}
			else {
				response.add("Success");
			}
		}
		return response;

	}

	public void close() {
		mmsScsmClientReceiver.close();
	}

	public void disconnect() {
		mmsScsmClientReceiver.disconnect();
	}

	public AlternateAccess checkArray(String objRef) {
		int openbracket = objRef.indexOf('(');

		if (openbracket != -1) {// is an array
			int closebracket = objRef.indexOf(')');
			List<AlternateAccess.SubChoice> listOfSeq = new ArrayList<AlternateAccess.SubChoice>();
			AlternateAccess subArrayEle = null;

			// for index
			BerInteger index = new BerInteger(Integer.parseInt(objRef.substring(openbracket + 1, objRef.indexOf(')'))));

			if (closebracket + 1 != objRef.length()) {
				// TODO ok to just use the remainder of the string, or does it
				// have to have the FC inserted?
				try {
					BerVisibleString arrayEle = new BerVisibleString(objRef
							.substring(closebracket + 2, objRef.length()).replace('.', '$'));
					// component name is stored in an AlternateAccess
					List<AlternateAccess.SubChoice> subSeqOf = new ArrayList<AlternateAccess.SubChoice>();
					subSeqOf.add(new AlternateAccess.SubChoice(null, arrayEle, null, null, null, null));
					subArrayEle = new AlternateAccess(subSeqOf);

					SubChoice_accessSelection accSel = new SubChoice_accessSelection(null, index, null, null);
					SubSeq_selectAlternateAccess selectAltAcc = new SubSeq_selectAlternateAccess(accSel, subArrayEle);

					listOfSeq.add(new AlternateAccess.SubChoice(selectAltAcc, null, null, null, null, null));

				} catch (UnsupportedEncodingException e) {
					throw new IllegalAccessError();
				}
			}
			else {
				listOfSeq.add(new AlternateAccess.SubChoice(null, null, index, null, null, null));
			}

			return new AlternateAccess(listOfSeq);
		}

		return null;
	}

	public void getRCBValues(ReportControlBlock rcb) throws ServiceError {
		MmsPdu mmsReq = constructGetRCBValuesRequest(rcb);
		MmsPdu mmsRes = encodeWriteReadDecode(mmsReq);
		decodeGetRCBValuesResponse(mmsRes, rcb);
	}

	private MmsPdu constructGetRCBValuesRequest(ReportControlBlock rcb) {
		ObjectReference objRef = rcb.getReference();
		ObjectName objectName = null;
		ObjectName.SubSeq_domain_specific subSeq = null;
		String funcCon = "";

		if (rcb instanceof UnbufferedReportContrlBlock) {
			funcCon = "$RP$";
		}
		else if (rcb instanceof BufferedReportControlBlock) {
			funcCon = "$BR$";
		}

		try {
			subSeq = new ObjectName.SubSeq_domain_specific(new BerVisibleString(objRef.get(0)), new BerVisibleString(
					objRef.get(1) + funcCon + objRef.get(2)));
		} catch (UnsupportedEncodingException e) {
			throw new IllegalAccessError();
		}
		objectName = new ObjectName(null, subSeq, null);

		VariableSpecification varSpec = new VariableSpecification(objectName, null, null, null, null);

		ListOfVariableSeq seq = new ListOfVariableSeq(varSpec, null);

		List<ListOfVariableSeq> listOfVariables = new ArrayList<ListOfVariableSeq>(1);
		listOfVariables.add(seq);

		VariableAccessSpecification varAccessSpec = new VariableAccessSpecification(new SubSeqOf_listOfVariable(
				listOfVariables), null);

		ReadRequest readRequest = new ReadRequest(new BerBoolean(false), varAccessSpec);

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(null, readRequest, null, null,
				null, null, null);

		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(getInvokeID()),
				confirmedServiceRequest);
		MmsPdu mms = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);
		return mms;
	}

	private void decodeGetRCBValuesResponse(MmsPdu responsePdu, ReportControlBlock rcb) throws ServiceError {
		// data in the msg should correspond to order of children
		if (responsePdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding GetDataValuesReponsePdu");
		}

		ConfirmedResponsePdu confirmedResponsePdu = responsePdu.confirmedResponsePdu;
		checkInvokeID(confirmedResponsePdu, invokeID);

		ConfirmedServiceResponse confirmedServiceResponse = confirmedResponsePdu.confirmedServiceResponse;
		if (confirmedServiceResponse.read == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"Error decoding GetDataValuesReponsePdu");
		}

		ReadResponse readResponse = confirmedServiceResponse.read;
		List<AccessResult> listOfAccessResults = readResponse.listOfAccessResult.seqOf;

		if (listOfAccessResults.size() != 1) {
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "Multiple results received.");
		}

		AccessResult accRes = listOfAccessResults.iterator().next();
		if (accRes.failure == null) {

			List<Data> results = accRes.structure.seqOf;
			if (results.size() != rcb.getChildren().size()) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "Too many results received");
			}

			int i = 0;

			for (ModelNode child : rcb.getChildren()) {
				// since convertMmsDataToDataContainer processes structures
				// based on the FunctionalConstraint
				// it does not work here. Therefore need to convert each Data
				// separately
				MmsAsn1ComplexTypeConverter.convertMmsDataToDataContainer(child, null, results.get(i++), mmsFCs);
			}

		}
		else {
			throw new ServiceError(mmsDataAccessErrorToServiceError(accRes.failure),
					"decodeGetDataValuesResponse: DataAccessError received.");
		}
	}

	public void setRCBValues(ReportControlBlock rcb) throws ServiceError {
		// sent for each value of the report control block

		BOOLEAN rptEna = (BOOLEAN) rcb.getChild("RptEna");
		boolean oldValue = rptEna.getValue();
		if (oldValue == true) {
			mmsScsmClientReceiver.enableReportQueueing();
		}
		else {
			mmsScsmClientReceiver.disableReportQueueing();
		}

		// need to send RptEna = false first - otherwise cannot change values
		rptEna.setValue(false);

		MmsPdu mmsReq = constructSetRCBValuesRequest(rptEna, rcb);
		MmsPdu mmsRes = encodeWriteReadDecode(mmsReq);
		handleSetRCBValuesResponse(mmsRes);

		for (ModelNode child : rcb.getChildren()) {
			// these appear to only be set by the server
			if (!child.getNodeName().equals("SqNum") && !child.getNodeName().equals("ConfRev")
					&& !child.getNodeName().equals("RptEna")) {

				mmsReq = constructSetRCBValuesRequest(child, rcb);
				mmsRes = encodeWriteReadDecode(mmsReq);
				handleSetRCBValuesResponse(mmsRes);

			}
		}

		if (oldValue == true) {
			rptEna.setValue(true);
			mmsReq = constructSetRCBValuesRequest(rptEna, rcb);
			mmsRes = encodeWriteReadDecode(mmsReq);
			handleSetRCBValuesResponse(mmsRes);
		}

	}

	private MmsPdu constructSetRCBValuesRequest(ModelNode node, ReportControlBlock rcb) throws ServiceError {
		ObjectReference objRef = rcb.getReference();
		ObjectName objectName = null;
		ObjectName.SubSeq_domain_specific subSeq = null;
		String funcCon = "";

		if (rcb instanceof UnbufferedReportContrlBlock) {
			funcCon = "$RP$";
		}
		else if (rcb instanceof BufferedReportControlBlock) {
			funcCon = "$BR$";
		}

		try {
			subSeq = new ObjectName.SubSeq_domain_specific(new BerVisibleString(objRef.get(0)), new BerVisibleString(
					objRef.get(1) + funcCon + objRef.get(2) + "$" + node.getNodeName()));
		} catch (UnsupportedEncodingException e) {
			throw new IllegalAccessError();
		}
		objectName = new ObjectName(null, subSeq, null);

		VariableSpecification varSpec = new VariableSpecification(objectName, null, null, null, null);

		List<ListOfVariableSeq> listOfVariable = new LinkedList<ListOfVariableSeq>();
		ListOfVariableSeq varSeq = new ListOfVariableSeq(varSpec, null);
		listOfVariable.add(varSeq);

		VariableAccessSpecification variableAccessSpec = new VariableAccessSpecification(new SubSeqOf_listOfVariable(
				listOfVariable), null);

		List<Data> listOfData = new LinkedList<Data>();
		Data mmsData = MmsAsn1ComplexTypeConverter.convertModelNodeToMmsData(node);
		listOfData.add(mmsData);

		WriteRequest writeRequest = new WriteRequest(variableAccessSpec, new SubSeqOf_listOfData(listOfData));

		ConfirmedServiceRequest confirmedServiceRequest = new ConfirmedServiceRequest(null, null, writeRequest, null,
				null, null, null);

		ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu(new BerInteger(getInvokeID()),
				confirmedServiceRequest);

		MmsPdu mms = new MmsPdu(confirmedRequestPdu, null, null, null, null, null);

		return mms;
	}

	private void handleSetRCBValuesResponse(MmsPdu responsePdu) throws ServiceError {
		testForErrorResponse(responsePdu);

		if (responsePdu.confirmedResponsePdu == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"handleSetDataValuesResponse: improper response");
		}

		ConfirmedResponsePdu confRes = responsePdu.confirmedResponsePdu;

		if (confRes.confirmedServiceResponse == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"handleSetDataValuesResponse: improper response");
		}

		ConfirmedServiceResponse confSerRes = confRes.confirmedServiceResponse;

		if (confSerRes.write == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"handleSetDataValuesResponse: improper response");
		}

		WriteResponse response = confSerRes.write;

		// only one?
		WriteResponse.SubChoice subChoice = response.seqOf.get(0);

		if (subChoice.failure != null) {
			throw new ServiceError(ServiceError.DATA_ACCESS_ERROR);
		}

	}

	public Report getReport() throws ServiceError {
		MmsPdu mmsPdu = null;
		try {
			mmsPdu = incomingReports.take();
		} catch (InterruptedException e) {
		}

		if (mmsPdu.unconfirmedPDU == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"getReport: Error decoding server response");
		}

		UnconfirmedPDU unconfirmedRes = mmsPdu.unconfirmedPDU;

		if (unconfirmedRes.unconfirmedService == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"getReport: Error decoding server response");
		}

		UnconfirmedService unconfirmedServ = unconfirmedRes.unconfirmedService;

		if (unconfirmedServ.informationReport == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"getReport: Error decoding server response");
		}

		return processReport(unconfirmedServ.informationReport);
	}

	public Report getReport(int timeout) throws ServiceError {
		MmsPdu mmsPdu = null;
		try {
			mmsPdu = incomingReports.poll(timeout, TimeUnit.MILLISECONDS);
			if (mmsPdu == null) {
				return null;
			}
		} catch (InterruptedException e) {
		}
		if (mmsPdu.unconfirmedPDU == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"getReport: Error decoding server response");
		}

		UnconfirmedPDU unconfirmedRes = mmsPdu.unconfirmedPDU;

		if (unconfirmedRes.unconfirmedService == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"getReport: Error decoding server response");
		}

		UnconfirmedService unconfirmedServ = unconfirmedRes.unconfirmedService;

		if (unconfirmedServ.informationReport == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"getReport: Error decoding server response");
		}

		return processReport(unconfirmedServ.informationReport);
	}

	/**
	 * See iec61850-8-1 Table 64 for the order of the report information To get
	 * the updated values, use the getDataSet method for reports
	 * 
	 * @param rpt
	 * @return
	 */
	private Report processReport(InformationReport rpt) throws ServiceError {
		List<AccessResult> listRes = rpt.listOfAccessResult.seqOf;
		int index = 0;

		if (listRes.get(index).visible_string == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"processReport: report does not contain RptID");
		}

		String rptID = listRes.get(index++).visible_string.toString();

		if (listRes.get(index).bit_string == null) {
			throw new ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
					"processReport: report does not contain OptFlds");
		}

		OptFields optFlds = new OptFields(new ObjectReference("temp"));
		optFlds.setValue(listRes.get(index++).bit_string.bitString);

		Report report = new Report();

		if (optFlds.isSeqNum()) {
			report.setSqNum((int) listRes.get(index++).unsigned.val);
		}

		if (optFlds.isTimeStamp()) {
			Timestamp time = new Timestamp(new ObjectReference(rptID), null, "", new Date(), false, false);
			MmsAsn1PrimitiveTypeConverter.toTimestamp(time, listRes.get(index++).binary_time.octetString);
			report.setTimeOfEntry(time);
		}

		if (optFlds.isDataSet()) {
			report.setDataSetRef(listRes.get(index++).visible_string.toString().replace('$', '.'));
		}

		if (optFlds.isBufOvfl()) {
			report.setBufOvfl(listRes.get(index++).boolean_.val);
		}

		if (optFlds.isEntryId()) {
			// TODO change report.EntryId to OctetString
			// report.setEntryId(listRes.get(index++).integer.val);
			byte[] b = listRes.get(index++).octet_string.octetString;
			long val = 0;
			for (int i = 0; i < b.length; i++) {
				val += (b[i] & 0xff) << (8 * i);
			}
			report.setEntryId(val);
		}

		if (optFlds.isConfigRef()) {
			// TODO change report.convRev to unsigned instead of integer?
			report.setConvRev(listRes.get(index++).unsigned.val);
		}

		if (optFlds.isSegmentation()) {
			report.setSubSqNum((int) listRes.get(index++).unsigned.val);
			report.setMoreSegmentsFollow(listRes.get(index++).boolean_.val);
		}

		byte[] inclusion = listRes.get(index++).bit_string.bitString;
		report.setInclusionBitString(inclusion);

		if (optFlds.isDataRef()) {
			// this is just to move the index to the right place
			// The next part will process the changes to the values
			// without the dataRefs
			for (int i = 0; i < inclusion.length * 8; i++) {
				if ((inclusion[i / 8] & (1 << (7 - i % 8))) == (1 << (7 - i % 8))) {
					index++;
				}
			}
		}

		// updating of data set copy - original stays the same
		DataSet ds = model.getDataSet(report.getDataSetRef()).copy();
		int shiftNum = 0;
		int numChanged = 0;

		for (ModelNode child : ds.getChildren()) {
			if ((inclusion[shiftNum / 8] & (1 << (7 - shiftNum % 8))) == (1 << (7 - shiftNum % 8))) {

				AccessResult accessRes = listRes.get(index++);
				MmsAsn1ComplexTypeConverter.convertMmsDataToDataContainer(child, child.getFunctionalConstraint(),
						new Data(accessRes.array, accessRes.structure, accessRes.boolean_, accessRes.bit_string,
								accessRes.integer, accessRes.unsigned, accessRes.floating_point,
								accessRes.octet_string, accessRes.visible_string, accessRes.generalized_time,
								accessRes.binary_time, accessRes.bcd, accessRes.booleanArray, accessRes.mMSString,
								accessRes.utc_time), mmsFCs);
				numChanged++;
			}
			shiftNum++;
		}

		report.setDataSet(ds);

		if (optFlds.isReasonCode()) {
			List<ReasonCode> reasonCodes = new ArrayList<ReasonCode>();
			for (int i = 0; i < numChanged; i++) {

				byte[] reason = listRes.get(index++).bit_string.bitString;

				if ((reason[0] & 0x02) == 0x02) {
					reasonCodes.add(ReasonCode.DCHG);
				}
				else if ((reason[0] & 0x04) == 0x04) {
					reasonCodes.add(ReasonCode.QCHG);
				}
				else if ((reason[0] & 0x08) == 0x08) {
					reasonCodes.add(ReasonCode.DUPD);
				}
				else if ((reason[0] & 0x10) == 0x10) {
					reasonCodes.add(ReasonCode.INTEGRITY);
				}
				else if ((reason[0] & 0x20) == 0x20) {
					reasonCodes.add(ReasonCode.GI);
				}
				else if ((reason[0] & 0x40) == 0x40) {
					reasonCodes.add(ReasonCode.APPTRIGGER);
				}
			}
		}

		return report;
	}
	//
	// public void close() {
	// if (pConnection != null) {
	// pConnection.disconnect();
	// }
	// }
	//
	// public List<String> getLogicalNodeDirectory(String lnRef, ACSIClass
	// acsiClass) throws ServiceError {
	// if (lnRef == null || !new ObjectReference(lnRef).isLNRef()) {
	// return null;
	// }
	//
	// List<String> dataNames = new LinkedList<String>();
	// IModelNode logicalNode = findOrRetrieveNode(lnRef);
	// if (logicalNode != null && logicalNode instanceof LogicalNode) {
	//
	// if (acsiClass == ACSIClass.DATA_SET) {
	// // retrieve DS list from server to get any configured dataSets
	// // as well
	// retrieveDataSetList((LogicalNode) logicalNode);
	// List<DataSet> lnDataSets = ((LogicalNode) logicalNode).getDataSets();
	// for (DataSet ds : lnDataSets) {
	// dataNames.add(ds.getNodeName());
	// }// non persistent DS are not included
	// }
	//
	// else {// data object
	//
	// if (!logicalNode.hasChildren()) {
	//
	// retrieveLogicalNodeDirectory((LogicalNode) logicalNode);
	// logicalNode = findOrRetrieveNode(lnRef);
	//
	// }
	//
	// for (IModelNode child : logicalNode) {
	// switch (child.getObjectType()) {
	// case DATA_OBJECT:
	// case ARRAY_OF_DO:
	// if (acsiClass == ACSIClass.DATA_OBJECT) {
	// dataNames.add(child.getNodeName());
	// }
	// break;
	// }
	// }
	//
	// }
	// }
	//
	// if (dataNames.size() == 0) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "No Objects found in LogicalNode. Logical Node should contain at least one DO.");
	// }
	// return dataNames;
	//
	// }
	//
	// protected void retrieveDataSetList(LogicalNode lnNode) throws
	// ServiceError {
	// // call getNameList using ld device name
	// LogicalDevice ldNode =
	// findOrRetrieveLogicalDevice(lnNode.getReference().getLDName());
	//
	// if (ldNode != null) {
	// // in case it exceeds the maximum mmsPdu size
	// String continueAfterRef = "";
	// do {
	// MmsPdu requestPdu =
	// constructGetLogicalNodeDataSetRequest(ldNode.getReference().toString(),
	// "");
	// // getParentRef() + "/"
	// // + lnNode.getNodeName(), "");// want LD/LN
	// MmsPdu responsePdu = encodeWriteReadDecode(requestPdu);
	// continueAfterRef = decodeGetDataSetListResponse(responsePdu, ldNode);
	//
	// } while (continueAfterRef != "");
	// }
	// else {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "retrieveDataSetList: Logical device not found");
	// }
	//
	// }
	//
	// private MmsPdu constructGetLogicalNodeDataSetRequest(String ldRef, String
	// continueAfter) throws ServiceError {
	// ObjectClass objectClass = new ObjectClass();
	// objectClass.selectBasicObjectClass((long)
	// MMSConstants.NAMED_VARIABLE_LIST);
	// GetNameListRequest getNameListRequest = new GetNameListRequest();
	// getNameListRequest.setObjectClass(objectClass);
	//
	// ObjectScopeChoiceType objectScopeChoiceType = new
	// ObjectScopeChoiceType();
	//
	// objectScopeChoiceType.selectDomainSpecific(new Identifier(ldRef));
	// getNameListRequest.setObjectScope(objectScopeChoiceType);
	//
	// ConfirmedServiceRequest confirmedServiceRequest = new
	// ConfirmedServiceRequest();
	// confirmedServiceRequest.selectGetNameList(getNameListRequest);
	// ConfirmedRequestPdu confirmedRequestPdu = new ConfirmedRequestPdu();
	// confirmedRequestPdu.setInvokeID(new
	// Unsigned32(Long.valueOf(getInvokeID())));
	// confirmedRequestPdu.setConfirmedServiceRequest(confirmedServiceRequest);
	// MmsPdu confirmedRequestMMSPdu = new MmsPdu();
	// confirmedRequestMMSPdu.selectConfirmedRequestPdu(confirmedRequestPdu);
	//
	// if (continueAfter != "") {
	// getNameListRequest.setContinueAfter(new Identifier(continueAfter));
	// }
	//
	// return confirmedRequestMMSPdu;
	//
	// }
	//
	// private String decodeGetDataSetListResponse(MmsPdu responsePdu,
	// LogicalDevice ldNode) throws ServiceError {
	// testForErrorResponse(responsePdu);
	//
	// if (!responsePdu.isConfirmedResponsePduSelected()) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "decodeGetDataSetListResponse: Error decoding server response");
	// }
	//
	// ConfirmedResponsePdu confirmedResponsePdu =
	// responsePdu.getConfirmedResponsePdu();
	// checkInvokeID(confirmedResponsePdu);
	//
	// ConfirmedServiceResponse confirmedServiceResponse =
	// confirmedResponsePdu.getConfirmedServiceResponse();
	// if (!confirmedServiceResponse.isGetNameListSelected()) {
	// throw new
	// ServiceError(ServiceError.FAILED_DUE_TO_COMMUNICATIONS_CONSTRAINT,
	// "decodeGetDataSetListResponse: Error decoding server response");
	// }
	//
	// GetNameListResponse getNameListResponse =
	// confirmedServiceResponse.getGetNameList();
	//
	// Collection<Identifier> identifiers =
	// getNameListResponse.getListOfIdentifier();
	//
	// if (identifiers.size() == 0) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE,
	// "decodeGetDataSetListResponse: Instance not available");
	// }
	//
	// // begin getting list of data sets
	// Identifier identifier = null;
	// Iterator<Identifier> it = identifiers.iterator();
	//
	// String idString;
	// List<DataSet> lnDataSets;
	//
	// int firstDollar;
	// LogicalNode lnNode;
	// String dataSetName;
	//
	// while (it.hasNext()) {
	// boolean add = true;
	// identifier = it.next();
	// idString = identifier.getValue(); // format of:
	// // LnName$FC$dataSetName
	//
	// // does the data set already exist?
	// // if yes, do not replace in case it refers to DOs/DAs
	// firstDollar = idString.indexOf('$');
	//
	// if (firstDollar < 0) {
	// throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "");
	// }
	//
	// // find logical node
	// lnNode = ldNode.getLN(idString.substring(0, firstDollar));
	//
	// if (lnNode == null) {
	// throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
	// "decodeGetDataSetListResponse: can't find logical node");
	// }
	//
	// lnDataSets = lnNode.getDataSets();
	// dataSetName = idString.substring(idString.lastIndexOf('$') + 1,
	// idString.length());
	//
	// for (DataSet dsCheck : lnDataSets) {
	// if (dataSetName.equals(dsCheck.getNodeName())) {
	// add = false;
	// }
	// }
	// if (add) {
	// DataSet ds = new DataSet(ldNode.getNodeName() + "/" +
	// idString.replace('$', '.'));
	// lnNode.addDataSet(ds);
	// }
	// }
	//
	// if (getNameListResponse.getMoreFollows() != null &&
	// !getNameListResponse.getMoreFollows()) {
	// return "";
	// }
	// else {
	// return identifier.getValue();
	// }
	// }
	//
	// protected void retrieveLogicalNodeDirectory(LogicalNode ln) throws
	// ServiceError {
	// if (ln != null) {
	// // retrieves model structures incl. the names of data objects
	// LogicalDevice parent =
	// findOrRetrieveLogicalDevice(ln.getReference().getParentRef());
	// retrieveLogicalNodes(parent);// this gets the LNs, DOs and FCs
	// }
	// // retrieveDataDefinitions(ln);
	// }
	//
	// public List<String> getDataDirectory(String doRef) throws ServiceError {
	// IModelNode doNode = getDataDefinition(doRef);
	// List<String> names = new LinkedList<String>();
	// for (IModelNode child : doNode) {
	// names.add(child.getNodeName());
	// }
	// return names;
	// }
	//
	// public IModelNode getDataDefinition(String ref) throws ServiceError {
	//
	// IModelNode node = findOrRetrieveNode(ref);
	// if (node != null && (node.hasChildren() || node.getObjectType() ==
	// ObjectType.BASIC_DATA_ATTRIBUTE)) {
	// return node;
	// }
	//
	// // Node is not complete
	// ObjectReference oRef = new ObjectReference(ref);
	// LogicalDevice ld = model.getLD(oRef);
	// if (ld == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE);
	// }
	// LogicalNode ln = ld.getLN(oRef);
	// if (ln == null) {
	// throw new ServiceError(ServiceError.INSTANCE_NOT_AVAILABLE);
	// }
	// retrieveDataDefinitions(ln);
	// return model.findNode(ref);
	// }
	//

}
