package org.openiec61850.scsm.mms.client;

import java.io.ByteArrayInputStream;
import java.lang.IllegalStateException;
import java.io.IOException;
import java.net.SocketTimeoutException;
import java.nio.ByteBuffer;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.BlockingQueue;

import org.openiec61850.jmms.iso.acse.AcseAssociation;
import org.openiec61850.jmms.mms.asn1.ConfirmedRequestPdu;
import org.openiec61850.jmms.mms.asn1.MmsPdu;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class MmsScsmClientReceiver extends Thread {

	private static Logger logger = LoggerFactory.getLogger(MmsScsmClientReceiver.class);

	private AcseAssociation acseAssociation;

	private BlockingQueue<MmsPdu> incomingResponses;
	private BlockingQueue<MmsPdu> incomingReports;
	private AtomicInteger expectedResponses = new AtomicInteger(0);
		
	private volatile boolean queueReports = false;
	private volatile boolean stopped = false;

	public MmsScsmClientReceiver(AcseAssociation acseAssociation, BlockingQueue<MmsPdu> incomingResponses,
			BlockingQueue<MmsPdu> incomingReports) {
		this.acseAssociation = acseAssociation;
		this.incomingResponses = incomingResponses;
		this.incomingReports = incomingReports;
	}

	public void run() {
		try {
			ByteBuffer encodedResponsePdu = null;
			while (true) {
				try {
					encodedResponsePdu = acseAssociation.receive();
				} catch (SocketTimeoutException e) {
					if (stopped == false && expectedResponses.get() == 0) {
						continue;
					} else {
						shutdown();
						break;
					}
				}

				MmsPdu decodedResponsePdu = new MmsPdu();
				try {
					decodedResponsePdu.decode(
							new ByteArrayInputStream(encodedResponsePdu.array(), encodedResponsePdu.arrayOffset()
									+ encodedResponsePdu.position(), encodedResponsePdu.limit()
									- encodedResponsePdu.position()), null);
				} catch (IOException e) {
					logger.warn("unable to decode received message: " + e.getMessage());
					shutdown();
					break;
				}

				if (decodedResponsePdu.unconfirmedPDU != null) {
					if (queueReports == true) {
						try {
							incomingReports.put(decodedResponsePdu);
						} catch (InterruptedException e) {
						}
					}
					else {
						logger.debug("discarding report because reports are disabled");
					}
				}
				else {
					try {
						incomingResponses.put(decodedResponsePdu);
						if(expectedResponses.decrementAndGet() < 0) {
							logger.warn("unexpected incoming response");
							shutdown();
							break;
						}								
					} catch (InterruptedException e) {
					}
				}
			}
		} catch (IOException e) {
			if (stopped == false) {
				logger.info("server closed connection unexpectedly");
				shutdown();
			}
		} catch (IllegalStateException e) {
			logger.info("unexpected connection close");
			shutdown();
		}
	}

	public void enableReportQueueing() {
		queueReports = true;
	}

	public void disableReportQueueing() {
		queueReports = false;
	}

	public void setResponseExpected() {
		expectedResponses.incrementAndGet();
	}

	public void close() {
		stopped = true;
		try {
			acseAssociation.close();
		} catch (Throwable e) {
			// ignore
		}
	}

	public void disconnect() {
		stopped = true;
		try {
			acseAssociation.disconnect();
		} catch (Throwable e) {
			// ignore
		}
	}

	private void shutdown() {
		if (stopped == false) {
			disconnect();
		}
		try {
			incomingResponses.put(new MmsPdu(new ConfirmedRequestPdu(), null, null, null, null, null));
			incomingReports.put(new MmsPdu(new ConfirmedRequestPdu(), null, null, null, null, null));
		} catch (InterruptedException e1) {
		}
	}				
}
